function [newpart, wnew, xest, info] = softConAuxPF(k, xpartminus, wminus, yobv)
% Soft-constrained Auxiliary Particle Filter


global Q;
global R;

global sigcon;
global sigbound;
global sigvel;

sigbound = sigcon(1); % for road boundary
sigvel = sigcon(3); % for speed limit

% ====== Initialisation ===================================================
ct = tic; % overall calculation time;
[n, N] = size(xpartminus); % sample size

xest = zeros(n,1);


invQ = inv(Q);
rankQ = rank(Q);
cholQ = chol(Q);
detQ=det(Q);
invR=inv(R);

%==========================================================================

xpartplus = fDyn(k,xpartminus); % x_k+1 = f(x_k)

bound = gCon(xpartplus);


%==========================================================================



%==========================================================================
ot = tic; % optimisation time

% coef of the Gaussian distribution 
coefnorm = (2*pi)^(-rankQ/2)/sqrt(detQ);
adjust = coefnorm * ones(1,N);

xpartmode = xpartplus; % initialise the xplusmode 

% % uncomment the following line if using Matlab optimisation routine
%--------------------------------------------------------------------------
% options = optimoptions('fminunc','SpecifyObjectiveGradient',true,'Algorithm','trust-region','MaxIter',5,'MaxFunEvals',50,'Display','off');
%--------------------------------------------------------------------------

for ii = 1:N
    
    if sum(bound(:,ii) > 0) % Alternatively, set bound + 1 sigma > 0 to decide if the constraints are violated.
        
        % optimisation:
        fi = xpartplus(:,ii);   
        s0 = fi; % initial guess 
        
%--------------------------------------------------------------------------        
%       % uncomment the following two lines and comment out the third line if using Matlab optimisation routine
%         obj = @(s1)costFun(s1,fi);
%         s0_opt = fminunc(obj,s0,options);
        s0_opt = Quasi_Newton (s0, fi, invQ, rankQ);
%--------------------------------------------------------------------------      
        xpartmode(:,ii) = s0_opt; % mode that considers the soft-constraints   

        
        % Calculate the adjustment for auxiliary resampling.
        temp = (fi-s0_opt)'*invQ*(fi-s0_opt);
        adjust(ii) = coefnorm * exp(-0.5*temp);
% %         
%         figure(1)
%         plot([fi(1) xpartmode(1,ii)],[fi(2) xpartmode(2,ii)],'.-k');     
    else
%         figure(1)
%         plot(xpartplus(1,ii),xpartplus(2,ii),'.b');   
    end
       
end
% =========================================================================
timeOpt = toc(ot); % optimisation time;


% ===================== Weights modification ==============================


dif=ones(N,1)*yobv'-(hObv(xpartmode))';
temp= sum((dif*invR).*dif,2);
auxweight = coefnorm*exp(-0.5*temp');


weight = wminus.*auxweight.* adjust;


weightsum = sum(weight);
weight = weight/weightsum;

% =========================================================================

% ===================== Resampling ========================================

[modenew, wnew, indexnew] = resampling(xpartmode,weight);

% =========================================================================


% % =========================================================================
% % draw new sample from mode
% 
% xpartnew = xpartmode + chol(Q)*randn(2,N);
% % =========================================================================


ratio = ones(1,N); % correction for transferred mode
xpartnew = zeros(n,N);
logratio = zeros(1,N);

for ii = 1:N
   
    idx = indexnew(ii);
    
    wnoise = cholQ*randn(rankQ,1);
    xpartnew(:,ii) = modenew(:,ii) + wnoise;
        
    s1 = xpartnew(:,ii);
    fi = xpartplus(:,idx);
    UTmodepart =modenew(:,ii);
    
    if fi ~= UTmodepart

        temp1 = (s1-fi)'*invQ*(s1-fi);
        temp2 = (s1-UTmodepart)'*invQ*(s1-UTmodepart);
        temp = temp1 - temp2;
        correction = exp(-temp/2);

        ratio(ii) = correction;
        logratio(ii) = -temp/2;  
    end
    
end

% ======================== soft constraint update p(c|x) ==================
gVal = gCon(xpartnew);

proInBound1 = myexpcdf(-gVal(1,:),sigbound);
proInBound2 = myexpcdf(-gVal(2,:),sigbound);
proInBound3 = myexpcdf(-gVal(3,:),sigvel);

logProIB1 = log(proInBound1);
logProIB2 = log(proInBound2);
logProIB3 = log(proInBound3);

% incon = proInBound1.*proInBound2.*proInBound3;
logIncon = logProIB1 + logProIB2 + logProIB3;


% =========================================================================

% ===============  new measurement update =================================

dif=ones(N,1)*yobv'-(hObv(xpartnew))';
temp1= sum((dif*invR).*dif,2);
update1 = coefnorm*exp(-0.5*temp1');


update2 = auxweight(indexnew);



% update = update1./update2;
logupdate = log(update1)-log(update2);

% =========================================================================


% wplus = wnew .* update .*incon.* ratio./ adjust(indexnew);
logwplus = log(wnew) + logupdate + logIncon + logratio - log(adjust(indexnew));


maxlogwplus = max(logwplus);

logwplus = logwplus - maxlogwplus;
wplus = exp(logwplus);

wplus = wplus /sum(wplus);
    
ess = ESS(wplus);

[xpartnew, wplus, ~] = resampling(xpartnew,wplus);

time = toc(ct); % overall calculation time;

% ============= Outputs ==================================================

xest = sum(xpartnew.*(ones(n,1)*wplus), 2);
newpart = xpartnew;
info.ess = ess;
info.fitness = 0; % not used
info.time = time;
info.timeOpt = timeOpt;

% =======================================================================

end



function ess = ESS(w)

sumw2 = w*w';
ess = 1/sumw2;

end



function [J, gradJ] = costFun(s1,fi)

global Q;

global sigbound;
global sigvel;

p3 = 5e-5; 
p2 = -0.004;
p1 = -0.2;
p0 = 125;



rnkQ = rank(Q);

invQ = inv(Q);

J0 = 0.5*(s1 - fi)'*invQ*(s1 - fi);


gVal = gCon(s1);

g1 = gVal(1);
g2 = gVal(2);
g3 = gVal(3);


if g1 <= 0
    Jcon1 = 0;
else
    Jcon1 = sigbound*g1;
end

if g2 <= 0
    Jcon2 = 0;
else
    Jcon2 = sigbound*g2;
end

if g3 <= 0
    Jcon3 = 0;
else
    Jcon3 = sigvel*g3;
end

J = J0 + Jcon1 + Jcon2 + Jcon3;


if nargout  > 1
    gradJ0 = (s1 - fi)'*invQ;
    
    if g1 < 0
        gradGcon1 = [0 0 0 0];
    else
        gradGcon1 = [0 1 0 0] - (3*p3*([1 0 0 0]*s1)^2 + 2*p2*([1 0 0 0]*s1) + p1)*[1 0 0 0];
    end
    
    if g2 < 0
        gradGcon2 = [0 0 0 0];
    else
        gradGcon2 = -([0 1 0 0] - (3*p3*([1 0 0 0]*s1)^2 + 2*p2*([1 0 0 0]*s1) + p1)*[1 0 0 0]);
    end
    
   
    bigQ = [zeros(2) zeros(2,2); zeros(2,2) eye(2,2)];

    if g3 <=0 
        gradGcon3 = [0 0 0 0];
    else
        gradGcon3 = 1/2 * (s1'*bigQ*s1)^(-1/2) * (2*s1'*bigQ);
    end
    
    gradJ1 = sigbound * gradGcon1;
    gradJ2 = sigbound * gradGcon2;
    gradJ3 = sigvel * gradGcon3;

    
    gradJ = gradJ0 + gradJ1 + gradJ2 + gradJ3;
    
end


end






